import zmq
import yaml
from time import sleep
from nested_dict import nested_dict


def merge(a, b, path=None):
    "merges b into a"
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


class zmqController:
    def __init__(self, ip, port, fname="configs/init.yaml"):
        context = zmq.Context()
        self.ip = ip
        self.port = port
        self.socket = context.socket(zmq.REQ)
        self.socket.connect("tcp://"+str(ip)+":"+str(port))
        with open(fname) as fin:
            self.yamlConfig = yaml.safe_load(fin)

    def reset(self):
        self.socket.close()
        context = zmq.Context()
        self.socket = context.socket(zmq.REQ)
        self.socket.connect("tcp://"+str(self.ip)+":"+str(self.port))

    def update_yamlConfig(self, fname="", yamlNode=None):
        if yamlNode:
            config = yamlNode
        elif fname:
            with open(fname) as fin:
                config = yaml.safe_load(fin)
        else:
            print("ERROR in %s" % (__name__))
        merge(self.yamlConfig, config)

    def initialize(self, fname="", yamlNode=None):
        self.socket.send_string("initialize", zmq.SNDMORE)
        if yamlNode:
            config = yamlNode
        elif fname:
            with open(fname) as fin:
                config = yaml.safe_load(fin)
        else:
            config = self.yamlConfig
        self.socket.send_string(yaml.dump(config))
        rep = self.socket.recv()
        print("returned status (from init) = %s" % rep)
        # return rep

    def configure(self, fname="", yamlNode=None):
        self.socket.send_string("configure", zmq.SNDMORE)
        if yamlNode:
            config = yamlNode
        elif fname:
            with open(fname) as fin:
                config = yaml.safe_load(fin)
        else:
            config = self.yamlConfig
        self.socket.send_string(yaml.dump(config))
        rep = self.socket.recv_string()
        print("returned status (from config) = %s" % rep)


class i2cController(zmqController):
    def __init__(self, ip, port, fname="configs/init.yaml"):
        super(i2cController, self).__init__(ip, port, fname)
        self.maskedDetIds = []

    def read_config(self, yamlNode=None):
        # only for I2C server
        self.socket.send_string("read", zmq.SNDMORE)
        # rep = self.socket.recv_string()
        if yamlNode:
            self.socket.send_string(yaml.dump(yamlNode))
        else:
            self.socket.send_string("")
        yamlread = yaml.safe_load(self.socket.recv_string())
        return(yamlread)

    def read_pwr(self):
        # only valid for hexaboard/trophy systems
        self.socket.send_string("read_pwr")
        rep = self.socket.recv_string()
        pwr = yaml.safe_load(rep)
        return(pwr)

    def resettdc(self):
        self.socket.send_string("resettdc")
        rep = self.socket.recv_string()
        return(yaml.safe_load(rep))

    def measadc(self, yamlNode):
        # only valid for hexaboard/trophy systems
        self.socket.send_string("measadc", zmq.SNDMORE)
        # rep = self.socket.recv_string()
        # if rep.lower().find("ready")<0:
        #     print(rep)
        #     return
        if yamlNode:
            config = yamlNode
        else:
            config = self.yamlConfig
        self.socket.send_string(yaml.dump(config))
        rep = self.socket.recv_string()
        adc = yaml.safe_load(rep)
        return(adc)

    def addMaskedDetId(self, detid):
        self.maskedDetIds.append(detid)

    def rmMaskedDetId(self, detid):
        self.maskedDetIds.remove(detid)

    def configure_injection(self, injectedChannels, activate=0, gain=0, phase=None, calib_dac=0):
        nestedConf = nested_dict()
        for key in self.yamlConfig.keys():
            if key.find('roc_s') == 0:
                if calib_dac == -1:
                    nestedConf[key]['sc']['ReferenceVoltage']['all']['IntCtest'] = 0
                else:
                    nestedConf[key]['sc']['ReferenceVoltage']['all']['IntCtest'] = activate
                    nestedConf[key]['sc']['ReferenceVoltage']['all']['Calib'] = calib_dac
                if not None == phase:  # no default phase, we don't change when not set
                    nestedConf[key]['sc']['Top']['all']['phase_ck'] = phase
                for injectedChannel in injectedChannels:
                    nestedConf[key]['sc']['ch'][injectedChannel]['LowRange'] = 0
                    nestedConf[key]['sc']['ch'][injectedChannel]['HighRange'] = 0
                    if gain == 0:
                        nestedConf[key]['sc']['ch'][injectedChannel]['LowRange'] = activate
                    else:
                        nestedConf[key]['sc']['ch'][injectedChannel]['HighRange'] = activate
        self.configure(yamlNode=nestedConf.to_dict())
        if not None == phase:
            self.resettdc()  # Reset MasterTDCs


class daqController(zmqController):
    def start(self):
        status = "none"
        while status.lower().find("running") < 0:
            self.socket.send_string("start")
            status = self.socket.recv_string()
            print(status)

    def is_done(self):
        self.socket.send_string("status")
        status = self.socket.recv_string()
        if status.lower().find("configured") < 0:
            return False
        else:
            return True

    def delay_scan(self):
        # only for daq server to run a delay scan
        rep = ""
        while rep.lower().find("delay_scan_done") < 0:
            self.socket.send_string("delayscan")
            rep = self.socket.recv_string()
        print(rep)

    def stop(self):
        self.socket.send_string("stop")
        rep = self.socket.recv_string()
        print(rep)

    def enable_fast_commands(self, random=0, external=0, sequencer=0, ancillary=0):
        self.yamlConfig['daq']['l1a_enables']['random_l1a'] = random
        self.yamlConfig['daq']['l1a_enables']['external_l1as'] = external
        self.yamlConfig['daq']['l1a_enables']['block_sequencer'] = sequencer

    def l1a_generator_settings(self, name='A', enable=0x0, BX=0x10, length=43, flavor='L1A', prescale=0, followMode='DISABLE'):
        for gen in self.yamlConfig['daq']['l1a_generator_settings']:
            if gen['name'] == name:
                gen['BX'] = BX
                gen['enable'] = enable
                gen['length'] = length
                gen['flavor'] = flavor
                gen['prescale'] = prescale
                gen['followMode'] = followMode

    def l1a_settings(self, bx_spacing=43, external_debounced=0, ext_delay=0, prescale=0, log2_rand_bx_period=0):  # ,length=43
        self.yamlConfig['daq']['l1a_settings']['bx_spacing'] = bx_spacing
        self.yamlConfig['daq']['l1a_settings']['external_debounced'] = external_debounced
        # self.yamlConfig['daq']['l1a_settings']['length']              = length
        self.yamlConfig['daq']['l1a_settings']['ext_delay'] = ext_delay
        self.yamlConfig['daq']['l1a_settings']['prescale'] = prescale
        self.yamlConfig['daq']['l1a_settings']['log2_rand_bx_period'] = log2_rand_bx_period

    def ancillary_settings(self, bx=0x10, prescale=0, length=100):
        self.yamlConfig['daq']['ancillary_settings']['bx'] = bx
        self.yamlConfig['daq']['ancillary_settings']['prescale'] = prescale
        self.yamlConfig['daq']['ancillary_settings']['length'] = length



import yaml
import logging
from client import Client
from hgcal_utilities import hgcal_utilities


class i2cController:
    """
    Compatibility layer between the hgcroc-tools i2c client interface and
    scripts expecting the zmqController and i2cController client interface.
    """

    def __init__(self, ip, port, fname=None):
        """
        Initialize the client and class variables, and get description
        from the hexaboard

        Parameters
        ----------
        ip
            Address or hostname of the board to connect to
        port
            Port of the board expecting the connection
        fname
            Initial configuration to use unless overridden by a parameter
            supplied to a different method. Included for compatibility
        """
        self.logger = logging.getLogger("i2cController")
        self.client = Client(ip, port)
        # Get board description
        self.description = self.client.describe()

        if fname is not None:
            with open(fname, "r") as f:
                raw_config = yaml.safe_load(f)
            self.yamlConfig = self._convert(raw_config)
        else:
            self.logger.warning("No default config specified")

    def _convert(self, raw_dict):
        board_type = list(self.description.keys())[0]
        board_layout = self.description[board_type]
        rocs = board_layout.keys()
        # Trim off configuration parameters that are not part of hexaboard
        trimmed_config = {}
        for roc in rocs:
            if roc not in raw_dict.keys():
                self.logger.warning(f'ROC {roc} not in configuration. Make '
                                    f'sure configuration is for {board_type}')
            else:
                trimmed_config[roc] = raw_dict[roc]['sc']
        # Expand block entries defined by 'all' into dicts for each block,
        # then merge any specified blocks into this
        converted_dict = {}
        for roc in trimmed_config.keys():
            converted_dict[roc] = {}
            for subBlock, block_dict in trimmed_config[roc].items():
                if 'all' in block_dict.keys():
                    global_config = block_dict['all']
                    del block_dict['all']
                    blocks_with_all = {key: global_config for key in
                                       board_layout[roc][subBlock].keys()}
                    # Merge specified blocks into blocks_with_all
                    for block in block_dict.keys():
                        blocks_with_all[block] = \
                                hgcal_utilities.update_dict(
                                        blocks_with_all[block],
                                        block_dict[block])
                # Note: yaml will dump this dict with references by default
                converted_dict[roc][subBlock] = blocks_with_all

        return converted_dict

    def _chooseNotNone(self, config_file: str, config_dict: dict):
        """
        Returns whichever non-default configuration is not 'None'. If both are
        not 'None', returns the default_config. If both are not specified,
        prioritizes config_dict. This behaviour is taken from the zmqController
        """
        return_config = self.yamlConfig
        if config_file is None:
            if config_dict is not None:
                return_config = self._convert(config_dict)
        else:
            with open(config_file, "r") as f:
                raw_config = yaml.safe_load(f)
            return_config = self._convert(raw_config)

        return return_config

    def reset(self):
        self.client.reset()

    def update_yamlConfig(self, fname=None, yamlNode=None):
        """
        Merge a new configuration into the existing one

        Parameters
        ----------
        fname
            A path to a yaml file containing a configuration to be merged.
            Lower priority than yamlNode
        yamlNode
            A configuration dictionary. If fname is specified, this will be
            preferred. This behaviour is taken from the original zmqController
        """
        new_config = self._chooseNotNone(fname, yamlNode)
        if new_config == self.yamlConfig:
            self.logger.warning("No new configuration specified to update")
        else:
            self.yamlConfig = hgcal_utilities.update_dict(
                    self.yamlConfigconfig, new_config)

    def initialize(self, fname=None, yamlNode=None):
        config = self._chooseNotNone(fname, yamlNode)
        self.client.set(config)

    def configure(self, fname=None, yamlNode=None):
        config = self._chooseNotNone(fname, yamlNode)
        self.client.set(config)

    def read_config(self, yamlNode=None):
        if yamlNode is None:
            response = self.client.get()
        else:
            not_nulled = self._convert(yamlNode)
            response = self.client.get(not_nulled)

        return response

    def read_pwr(self):
        raise NotImplementedError("read_pwr not implemented")

    def resettdc(self):
        board_type = list(self.description.keys())[0]
        board_layout = self.description[board_type]
        raw_config = {roc: {'sc': {'MasterTdc': {'all': {'START_COUNTER': 0}}}}
                      for roc in board_layout.keys()}
        self.configure(yamlNode=raw_config)  # formats and sends dict
        raw_config = {roc: {'sc': {'MasterTdc': {'all': {'START_COUNTER': 1}}}}
                      for roc in board_layout.keys()}
        self.configure(yamlNode=raw_config)  # formats and sends dict

        return "masterTDCs reset."

    def measadc(self, yamlNode):
        raise NotImplementedError("measadc not implemented")

    def configure_injection(self, injectedChannels, activate=0,
                            gain=0, phase=None, calib_dac=0):
        board_type = list(self.description.keys())[0]
        board_layout = self.description[board_type]
        if gain == 0:
            LR = activate
            HR = 0
        else:
            LR = 0
            HR = activate
        config = {}
        for roc in board_layout.keys():
            if calib_dac == -1:
                config[roc] = {'ReferenceVoltage': {0: {'IntCtest': 0},
                                                    1: {'IntCtest': 0}}}
            else:
                config[roc] = {'ReferenceVoltage':
                               {0: {'IntCtest': activate},
                                1: {'IntCtest': activate}}}
                config[roc]['ReferenceVoltage'][0]['Calib'] = \
                    calib_dac
                config[roc]['ReferenceVoltage'][1]['Calib'] = \
                    calib_dac
            if phase is not None:
                config[roc]['Top'] = {0: {'phase_ck': phase}}
            config[roc]['ch'] = {}
            for channel in injectedChannels:
                config[roc]['ch'][channel] = {'LowRange': LR,
                                              'HighRange': HR}
        self.client.set(config)
        if phase is not None:
            self.resettdc()
